//-----------------------------------------------------------------------------
// Copyright (c) 2008-2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Dhpoware;

namespace XNA2GaussianBlur
{
    /// <summary>
    /// This Windows XNA application implements Gaussian blurring on a
    /// source texture image. The Gaussian blur is implemented as an HLSL
    /// effect. This demo performs the Gaussian blur on a bitmap image. However
    /// the Gaussian blur can be performed as a post process on a scene
    /// rendered to a render target texture. Gaussian blurs are commonly used
    /// in conjunction with more advanced post processing shader effects such
    /// as full screen specular bloom shaders.
    /// <para>
    /// A 2D Gaussian distribution is said to be circularly symmetric. This
    /// means that the Gaussian blur operation is linearly separable. Instead
    /// of applying the 2D filter kernel matrix to the image we can instead
    /// perform the Gaussian blur operation in two passes using a 1D filter
    /// kernel matrix. We perform a horizontal Gaussian blur pass on the image
    /// followed by a vertical Gaussian blur pass. This produces the same
    /// result as performing a single Gaussian blur pass on the image using the
    /// 2D filter kernel matrix. However performing the blur in two passes
    /// using 1D filter kernel matrices is much faster.
    /// </para>
    /// <para>
    /// For further details on Gaussian blurring see:
    /// http://en.wikipedia.org/wiki/Gaussian_blur
    /// </para>
    /// </summary>
    public class Demo : Microsoft.Xna.Framework.Game
    {
        private static void Main()
        {
            using (Demo demo = new Demo())
            {
                demo.Run();
            }
        }

        // WARNING:
        // If you change the BLUR_RADIUS you *MUST* also change the RADIUS
        // constant in GaussianBlur.fx. Both values *MUST* be the same.

        private const int BLUR_RADIUS = 7;
        private const float BLUR_AMOUNT = 2.0f;

        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;
        private SpriteFont spriteFont;
        private Texture2D texture;
        private RenderTarget2D renderTarget1;
        private RenderTarget2D renderTarget2;
        private Vector2 fontPos;
        private KeyboardState currentKeyboardState;
        private KeyboardState prevKeyboardState;
        private GaussianBlur gaussianBlur;
        private int windowWidth;
        private int windowHeight;
        private int renderTargetWidth;
        private int renderTargetHeight;
        private int frames;
        private int framesPerSecond;
        private TimeSpan elapsedTime = TimeSpan.Zero;
        private bool displayHelp;
        private bool enableGaussianBlur;
        
        public Demo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Window.Title = "XNA 4.0 Gaussian Blur";
            IsMouseVisible = true;
            IsFixedTimeStep = false;
        }
        
        protected override void Initialize()
        {
            // Setup the window to be a quarter the size of the desktop.

            windowWidth = GraphicsDevice.DisplayMode.Width / 2;
            windowHeight = GraphicsDevice.DisplayMode.Height / 2;

            // Setup frame buffer.
            
            graphics.SynchronizeWithVerticalRetrace = false;
            graphics.PreferredBackBufferWidth = windowWidth;
            graphics.PreferredBackBufferHeight = windowHeight;
            graphics.ApplyChanges();

            // Position the text.
            
            fontPos = new Vector2(1.0f, 1.0f);

            // Setup the initial input states.
            
            currentKeyboardState = Keyboard.GetState();
                        
            // Create the Gaussian blur filter kernel.
            
            gaussianBlur = new GaussianBlur(this);
            gaussianBlur.ComputeKernel(BLUR_RADIUS, BLUR_AMOUNT);
                        
            base.Initialize();
        }

        private void InitRenderTargets()
        {
            // Since we're performing a Gaussian blur on a texture image the
            // render targets are half the size of the source texture image.
            // This will help improve the blurring effect.

            renderTargetWidth = texture.Width / 2;
            renderTargetHeight = texture.Height / 2;

            renderTarget1 = new RenderTarget2D(GraphicsDevice,
                renderTargetWidth, renderTargetHeight, false,
                GraphicsDevice.PresentationParameters.BackBufferFormat,
                DepthFormat.None);

            renderTarget2 = new RenderTarget2D(GraphicsDevice,
                renderTargetWidth, renderTargetHeight, false,
                GraphicsDevice.PresentationParameters.BackBufferFormat,
                DepthFormat.None);

            // The texture offsets used by the Gaussian blur shader depends
            // on the dimensions of the render targets. The offsets need to be
            // recalculated whenever the render targets are recreated.

            gaussianBlur.ComputeOffsets(renderTargetWidth, renderTargetHeight);
        }

        private bool KeyJustPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && prevKeyboardState.IsKeyUp(key);
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>(@"Fonts\DemoFont");

            texture = Content.Load<Texture2D>(@"Textures\lena");

            InitRenderTargets();
        }

        private void ProcessKeyboard()
        {
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            if (KeyJustPressed(Keys.Escape))
                this.Exit();

            if (KeyJustPressed(Keys.Space))
                enableGaussianBlur = !enableGaussianBlur;

            if (KeyJustPressed(Keys.H))
                displayHelp = !displayHelp;

            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) ||
                currentKeyboardState.IsKeyDown(Keys.RightAlt))
            {
                if (KeyJustPressed(Keys.Enter))
                    ToggleFullScreen();
            }
        }

        private void ToggleFullScreen()
        {
            int newWidth = 0;
            int newHeight = 0;

            graphics.IsFullScreen = !graphics.IsFullScreen;

            if (graphics.IsFullScreen)
            {
                newWidth = GraphicsDevice.DisplayMode.Width;
                newHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                newWidth = windowWidth;
                newHeight = windowHeight;
            }

            graphics.PreferredBackBufferWidth = newWidth;
            graphics.PreferredBackBufferHeight = newHeight;
            graphics.ApplyChanges();
        }

        protected override void UnloadContent()
        {
            renderTarget1.Dispose();
            renderTarget1 = null;

            renderTarget2.Dispose();
            renderTarget2 = null;
        }

        protected override void Update(GameTime gameTime)
        {
            if (!IsActive)
                return;

            ProcessKeyboard();
            UpdateFrameRate(gameTime);

            base.Update(gameTime);
        }

        private void UpdateFrameRate(GameTime gameTime)
        {
            elapsedTime += gameTime.ElapsedGameTime;

            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                framesPerSecond = frames;
                frames = 0;
            }
        }

        private void IncrementFrameCounter()
        {
            ++frames;
        }

        private void DrawText()
        {
            StringBuilder buffer = new StringBuilder();

            if (displayHelp)
            {
                buffer.AppendLine("Press SPACE to enable/disable Gaussian blur");
                buffer.AppendLine("Press ALT and ENTER to toggle full screen");
                buffer.AppendLine("Press ESCAPE to exit");
                buffer.AppendLine();
                buffer.AppendLine("Press H to hide help");
            }
            else
            {
                buffer.AppendFormat("FPS: {0}\n", framesPerSecond);
                buffer.AppendLine();
                buffer.AppendFormat("Radius: {0}\n", gaussianBlur.Radius);
                buffer.AppendFormat("Sigma: {0}\n", gaussianBlur.Sigma.ToString("f2"));
                buffer.AppendLine();
                buffer.AppendLine("Press H to display help");
            }

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            spriteBatch.DrawString(spriteFont, buffer.ToString(), fontPos, Color.Yellow);
            spriteBatch.End();
        }

        protected override void Draw(GameTime gameTime)
        {
            if (!IsActive)
                return;

            if (enableGaussianBlur)
            {
                Texture2D result = gaussianBlur.PerformGaussianBlur(texture, renderTarget1, renderTarget2, spriteBatch);
                Rectangle rectangle = new Rectangle(0, 0, texture.Width, texture.Height);

                GraphicsDevice.Clear(Color.CornflowerBlue);

                spriteBatch.Begin();
                spriteBatch.Draw(result, rectangle, Color.White);
                spriteBatch.End();
            }
            else
            {
                GraphicsDevice.Clear(Color.CornflowerBlue);

                spriteBatch.Begin();
                spriteBatch.Draw(texture, new Rectangle(0, 0, texture.Width, texture.Height), Color.White);
                spriteBatch.End();
            }

            DrawText();
            base.Draw(gameTime);
            IncrementFrameCounter();
        }
    }
}